#ifndef _DT_BLKFX_PARAM_H_
#define _DT_BLKFX_PARAM_H_
/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

BlkFxParam : namespace to help with VST parameters used by DtBlkFx

History
	Date       Version    Programmer         Comments
	01/12/05    1.0        Darrell Tam		Created


This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/


#include <math.h>
#include <dt/NoteFreq.h>
#include <dt/misc_stuff.h>
#include "rfftw_float.h"

// namespace to hold stuff to do with vst params used in dt blkfx
namespace BlkFxParam
{

	// parameter values
	enum {
		// global params that apply to all fx
		MIX_BACK = 0,		// original mix back fraction & power match control
		DELAY,				// delay units is 1/4 * signature * bpm
		FFT_LEN,			// fft len is 
		OVERLAP,			// overlap fraction
		NUM_GLOBAL_PARAMS,

		// each set of fx has these params (add i*NUM_FX+NUM_GLOBAL)
		FX_FREQ_A = 0,
		FX_FREQ_B,
		FX_AMP,
		FX_TYPE,
		FX_VAL,
		NUM_FX_PARAMS,

		NUM_FX_SETS = 4,
		TOTAL_NUM = NUM_GLOBAL_PARAMS+NUM_FX_PARAMS*NUM_FX_SETS,

		// number of ticks per beat
		TICKS_PER_BEAT = 16,

		#ifdef STEREO
		AUDIO_CHANNELS = 2
		#elif MONO
		AUDIO_CHANNELS = 1
		#else
		#error define STEREO or MONO
		#endif
	};

	// get the parameter offset given an fx set
	inline int paramOffs(int fx_set) { return NUM_GLOBAL_PARAMS+fx_set*NUM_FX_PARAMS; }

	struct SplitParamNum
	{
		int param;		// original param number
		bool ok;		// true if param is in range
		int glob_param;	// global param number or -1 if effect set
		int fx_set;		// effect set or -1 if global
		int fx_param;	// effect param or -1 if global

		SplitParamNum(int param_num)
		{
			param = param_num;
			fx_param = -1;
			glob_param = -1;
			fx_set = -1;
			ok = false;
			if(param_num < 0 || param_num >= TOTAL_NUM) return;

			ok = true;
			fx_set = param_num-NUM_GLOBAL_PARAMS;
			if(fx_set < 0) {
				// global param
				fx_set = -1;
				glob_param = param_num;
			}
			else {
				// param belongs to effect set
				fx_param = fx_set%NUM_FX_PARAMS;
				fx_set = limit_range(fx_set/NUM_FX_PARAMS, 0, NUM_FX_SETS-1);
			}			
		}
	};

	// --- OVERLAP ---------------------------------------------------------------------------------------------
	inline bool getPwrMatch(float/*0..1*/mixback_param)
	// get pwr match part of mixback param
		{ return mixback_param <= 0.5f;	}

	inline float getMixBackFrac(float/*0..1*/mixback_param)
	// get the mixback fraction of the mixback param
	{
		return mixback_param<=0.5f ? mixback_param*2.0f : (1.0f-mixback_param)*(1.0f/0.499f);
	}

	inline float/*0..1*/getMixbackParam(float/*0..1*/mixback_frac, bool pwr_match)
	// combine pwr match and mixback fraction into overlap param
	{
		return pwr_match? 
			limit_range(mixback_frac*0.5f, 0.0f, 0.5f) :
			limit_range(1.0f-mixback_frac*0.499f, 0.501f, 1.0f);
	}

	// --- DELAY -----------------------------------------------------------------------------------------------
	inline float getDelayTicks(float/*0..1*/delay_param)
		{ return delay_param*255.0f; }

	// given delay in beats, return as param value
	inline float/*0..1*/getDelayParam(float delay_beats)
		{ return limit_range(delay_beats*(TICKS_PER_BEAT/255.0f), 0.0f, 1.0f); }

	// --- FFT_LEN ---------------------------------------------------------------------------------------------
	inline long/*plan index*/getPlan(float/*0..1*/fft_len_param)
	// given vst param for fft_len, return fft plan number
	{
		return limit_range(
			(int)(fft_len_param*255.0f)/16,
			0,
			N_FFT_BLK_SZ-1
		);
	}

	inline float/*0..1*/getFFTLenParam(int plan_index)
	// turn plan index into fft length
	{
		return limit_range(
			((float)plan_index+0.5f)*(16.0f/255.0f),
			0.0f, 1.0f
		);
	}

	// --- OVERLAP ---------------------------------------------------------------------------------------------
	inline bool getBlkSync(float/*0..1*/overlap_param)
	// get blk sync part of the overlap param
		{ return overlap_param > 0.5f; }

	inline float getOverlapPart(float/*0..1*/overlap_param)
	// get the overlap part of the overlap param
		{ return overlap_param > 0.5f ? 2.0f-overlap_param*2.0f : overlap_param*2.0f; }

	inline long getBlkShiftFwd(float/*0..1*/overlap_param, float fft_n)
	// return number of samples we will shift forward in the blk given the overlap_param
	// overlap param is divided into 2 parts:
	//   0 .. 0.5 = normal "blind" overlap where blks are not sync'd to anything
	//   0.5 .. 1 = blk sync mode where overlapping blks are sync'd to beats
	{
		float min_n = fft_n*.15f;
		if(min_n < 16.0f) min_n = 16.0f;
		return (long)lin_interp(
			getOverlapPart(overlap_param),
			(fft_n-10.0f),	// overlap=0: maximum shift
			min_n			// overlap=1: mimimum shift
		);
	}

	inline float/*0..1*/getOverlapParam(float/*0..1*/overlap_part, bool sync)
	//
	// combine overlap part and sync into overlap param
	//
	// overlap_part units are 0=approx 0%, 1=approx 85% (depends on blk sz)
	//
	{
		return sync? 
			limit_range((2.0f-overlap_part)*0.5f, 0.501f, 1.0f) :
			limit_range(overlap_part*0.5f, 0.0f, 0.499f);
	}

	// --- FX_FREQ_A/FX_FREQ_B --------------------------------------------------------------------
	// freq params correspond to exponential frequency which means that octaves are
	// linear and the param is actually a note offset from c0

	// number of notes spanned in freq param
	inline float noteSpan() { return 255.0f*0.5f; }

	// number of octaves spanned in freq param
	inline float octaveSpan() { return noteSpan()/12.0f; }

	// fraction of freq param corresponding to 1 octave
	inline float octaveFrac() { return 1.0f/octaveSpan(); }

	// fraction of freq param corresponding to 1 note
	inline float noteFrac() { return 1.0f/noteSpan(); }

	inline float/*freq hz*/getHzNonZero(float/*0..1*/param)
	// map the vst param to a freq (with no 0 hz)
	{
		return NoteOffsToHz(/*offs c0*/param*noteSpan());
	}

	inline float/*freq hz*/getHz(float/*0..1*/param)
	// if param is 0 then return 0, otherwise return "getFreqNonZero"
	{
		return param <= 0.0f ? 0.0f : getHzNonZero(param);
	}
	inline float/*bin*/getFFTBin(float/*0..1*/param, float fft_n, float samp_rate)
	//
	// given freq param value and samples per fft blk, return the corresponding bin index
	//
	{
		float freq_hz = getHz(param);
		return limit_range(fft_n*freq_hz/samp_rate, 0.0f, fft_n*0.5f);
	}


	// --- FX_TYPE --------------------------------------------------------------------------------
	inline long/*index*/getEffectType(float/*0..1*/param)
	{
		// align to renoise effect values, allow for up to 32 possible effects
		return (long)(param*255.0)/8;
	}

	// inverse function to return the effect type param given an index
	inline float/*param 0..1*/getEffectTypeInv(long index) { return (float)(index*8+4)/255.0f; }

	// --- FX_AMP ---------------------------------------------------------------------------------
	inline float /*dB*/ getEffectAmp(float/*0..1*/param)
	{
		return lin_interp(param, -60.0f, 40.0f);
	}
	// get 0dB point of amp param
	inline float getAmpParam0dB() { return .6f; }

	inline float /*multiply*/ getEffectAmpMult(float/*0..1*/param, bool mix_mode)
	// convert amp param into linear multiplier (i.e. param->dB->linear)
	// if mix_mode is true and amp is less than 0dB point, linear 
	{
		// check for mix_mode linear range
		if(mix_mode && param < getAmpParam0dB()) return param * (1.0f/getAmpParam0dB());

		// normal exponential range
		return param <= 0.0f? 0.0f: powf(10.0f, getEffectAmp(param)*0.05f);
	}

};

#endif
