/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

Top-level DtBlkFx gui

Holds the following panels:
    1x GlobalCtrl
    2x spectrogram
    4x FxCtrl


History
	Date       Version    Programmer         Comments
	16/2/03    1.0        Darrell Tam		Created


This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/

#include <math.h>
#include <stdlib.h>	
#include <stdio.h>
#include <valarray>
#include <dt/windows_support.h>

#include "BlkFxParam.h"
#include "DTBlkFx.hpp"
#include "FFTFx.h"

#include "Spectrogram.h"
#include "Gui.h"

#include "FxCtrl.h"

//#define DBG_FILE_NAME "c:\\gui.txt"
#include "Debug.h"

using namespace std;
using namespace VSTGUI;

enum
{
	// bitmaps
	// others
	kBackgroundW = 410,
	kBackgroundH = 410
};


//-------------------------------------------------------------------------------------------------
void Gui::generateAmpOverlay()
// generate exponential amplitude gfx into _amp_bm
// this code was written (instead of just using a normal bitmap) back when the whole gui was
// going to be scalable
{
	int width = _amp_bm->getWidth();
	int height = _amp_bm->getHeight();
	float fwidth = (float)width;
	float fheight = (float)height;

	// generate an exponential curve
	int zero_db = (int)(0.6f*fwidth);
	float xscale = 1.0f/zero_db;
	float yscale = 0.8f*fheight;
	CPoint p;
	for(p.x = 0; p.x < width; p.x++) {
		float top = yscale*(1-powf(20.0f, (float)p.x*xscale)/20.0f);

		int i_top = (int)floorf(top);
		for(p.y = 0; p.y < height; p.y++) {
			if(p.y < i_top) {
				_amp_bm->drawTransparent(p);
				continue;
			}
			unsigned alpha = 255;
			if(p.y == i_top) {
				// anti-alias
				float frac = (float)alpha*(1.0f-top+(float)i_top);
				alpha = (int)frac;
			}
			_amp_bm->drawPoint(p, 255, 0, 0, alpha);
		}
	}
}

//-------------------------------------------------------------------------------------------------
Gui::Gui (AudioEffect *effect) : AEffGUIEditor (effect)
{
	_ok = false;

	/*AEffGUIEditor*/frame = 0;

	/*AEffGUIEditor::rect*/
	rect.left   = 0;
	rect.top    = 0;
	rect.right  = kBackgroundW;
	rect.bottom = kBackgroundH;

	// build the colour gradient map for the spectrograms
	float col_vec[][4] = {
	/*   pos     R     G     B */
		{0.0f,  0.0f, 0.0f, 0.0f},
		{0.2f,  0.0f, 0.0f, 1.0f},
		{0.4f,  0.0f, 1.0f, 1.0f},
		{0.6f,  0.0f, 0.8f, 0.0f},
		{0.8f,  1.0f, 1.0f, 0.0f},
		{1.0f,  1.0f, 0.0f, 0.0f}
	};
	_col_map.generate(1000, col_vec, 6);
	for(int i = 0; i < 2; i++) _sgram[i].set(_col_map);

}

//-------------------------------------------------------------------------------------------------
Gui::~Gui ()
{
	DBG("Gui::~Gui" << VAR(this));
}

//-------------------------------------------------------------------------------------------------
long Gui::open (void *system_window)
// virtual, override AEffGUIEditor::open
// open the GUI window
//
{
	_ok = false;

	int i;
	AEffGUIEditor::open(system_window);

	// build pixel to freq mapping
	_pix_freq.init(
		400,								// width
		blkFx()->getSampleRate()			// sample freq in Hz
	);

	// build pixel to bin range mapping
	for(i = 0; i < BlkFxParam::AUDIO_CHANNELS; i++)
		_pix_bin[i].init(
			_pix_freq,						// 
			&(blkFx()->_chan[i].x1[0])		// FFT data buffer
		);

	// get version
	int version = getVstGuiVersion ();
	int verMaj = (version & 0xFF00) >> 16;
	int verMin = (version & 0x00FF);

	CRect r;
	CPoint p;

	// get bitmaps
	VstGuiRef<CBitmap> bg_bitmap;
	bg_bitmap.setNew(new CBitmap(10000));

	// create main window
	r(0, 0, kBackgroundW, kBackgroundH);
	frame = new CFrame (/*rectangle*/r, system_window, /*gui editor*/this);
	frame->setBackground (bg_bitmap);

	// 
	_freq_bg_bm.setNew(new CAlphaContext(frame));
	CopyResourceBitmapToAlphaContext(_freq_bg_bm, /*resource id*/10001);

	_freq_knob_bm.setNew(new CAlphaContext(frame));
	CopyResourceBitmapToAlphaContext(_freq_knob_bm, /*resource id*/10100);

	_fxval_knob_bm.setNew(new CAlphaContext(frame));
	CopyResourceBitmapToAlphaContext(_fxval_knob_bm, 10200);

	// place global controls
	p(0, 0);
	r(0, 0, _pix_freq.numPixels()+10, 15+13);
	if(_glob_ctrl.init(this, Offset(p, r))) frame->addView(&_glob_ctrl);

	// place spectrograms
	p(4, 28);
	for(i = 0; i < 2; i++) {
		Spectrogram& sgram = _sgram[i];
		int h = 93;				// height
		sgram.init(
			frame,				// CFrame
			p,					// offset point
			h,					// height
			&_pix_freq			// pixel to freq mapping
		);
		frame->addView(_sgram+i);

		// add a menu in stereo mode
		if(BlkFxParam::AUDIO_CHANNELS == 2) {
			sgram.menu = 
				new COptionMenu (
					CRect(0, 0, _pix_freq.numPixels(), h),	// rectangle
					NULL,				// CControlListener *
					-1,					// tag number
					NULL,				// background
					NULL,				// background on click
					k3DIn				// style
				);
			if(sgram.menu) {
				sgram.addView(sgram.menu);
				const char* menu[] = { "Input L+R", "Input Left", "Input Right", "Output L+R", "Output Left", "Output Right", "Paused" };
				for(int j = 0; j < 7; j++) sgram.menu->addEntry((char*)menu[j], j);
				float menu_init[] = {0, 3};
				sgram.menu->setValue(menu_init[i]);
			}
		}
		else {
			const char* name[] = { "Input", "Output" };
			_sgram[i].name = name[i];
		}

		CAlphaContext *sgram_context = sgram.getImage();
		bg_bitmap->draw(sgram_context, GetRect(sgram_context), TopLeft(sgram.client_rect));
		if(i == 1) {
			sgram_context->setFontColor(CColorRGB(255, 255, 255));
			sgram_context->setFillColor(CColorRGB(0, 0, 0));
			sgram_context->setFont(kNormalFont, /*sz*/0, kBoldFace);
			CRect txt_r(0, 10, sgram_context->getWidth(), 10+15);

			const char* tip[] = { "", "Drag with right mouse button on knobs & invisi-sliders for fine-control", "" };
			for(int l = 0; l < 3; l++) {
				OutlineDrawString(sgram_context, txt_r, tip[l]);
				txt_r.offset(0, 15);
			}
			txt_r.offset(0,5);
			OutlineDrawString(sgram_context, txt_r, "Compiled " __DATE__);
		}
		p.y += h;
	}
	GdiFlush();
	DBG("Gui::open, spectrograms ok");

	// how big each fx control panel is
	r(0, 0, _pix_freq.numPixels()+10, 45);

	// create amp overlay
	_amp_bm.setNew(new CAlphaContext(frame, r.width(), r.height()));
	if(_amp_bm && _amp_bm->ok()) generateAmpOverlay();
	DBG("Gui::open" << VAR(_amp_bm) << VAR(_amp_bm?_amp_bm->ok():false));

	// create fx ctrls
	p(0, 214);
	r(0, 0, _pix_freq.numPixels()+10, 48);
	for(i = 0; i < BlkFxParam::NUM_FX_SETS; i++) {
		if(_fx_ctrl[i].init(this, Offset(p, r))) frame->addView(_fx_ctrl+i);
		p.y += r.height();
	}

	// load params from the effect
	for(i = 0; i < BlkFxParam::TOTAL_NUM; i++) setParameter(i, blkFx()->getCurrParam(i));

	_ok = true;
	return true;
}


//-------------------------------------------------------------------------------------------------
static bool g_stereo_ch_menu_map[/*in-out*/2][/*menu*/6][/*ch*/2] = {
	{// in
		{true, true},		// L+R in
		{true, false},		// left in only
		{false, true},		// right in only
		{false, false},		// L+R out
		{false, false},		// left out only
		{false, false}		// right out only
	},
	{// out
		{false, false},		// L+R in
		{false, false},		// left in only
		{false, false},		// right in only
		{true, true},		// L+R out
		{true, false},		// left out only
		{false, true}		// right out only
	}
};

//-------------------------------------------------------------------------------------------------
void Gui::FFTDataRdy(
	int a			// 0=input data, 1=output data
)
{
	// note that there is a potential race if the gui is open, closed & re-opened during this
	// method but I think that it being closed & reopened so quickly as to cause a problem is
	// unlikely

	if(!_ok) return;

	int plan = blkFx()->_plan;
	long samp_pos = blkFx()->_blk_samp_abs;
	long time_fft_n = blkFx()->_time_fft_n;
	if(BlkFxParam::AUDIO_CHANNELS == 2) {
		// stereo, check which channels to display
		for(int sgram_i = 0; sgram_i < 2; sgram_i++) {
			Spectrogram& sgram = _sgram[sgram_i];
			if(!sgram.rdy()) continue;

			// convert menu number 0..6 to 0..2
			// input data corresponds to menu values 0..2, output data to 3..5
			int menu_val = (int)(sgram.menu->getValue());
			if(menu_val < 0 || menu_val >= 6) continue;
			
			// update spectrogram with audio data
			bool update = false;
			for(int ch = 0; ch < 2; ch++) {
				if(g_stereo_ch_menu_map[a][menu_val][ch]) {
					float pwr_scale = 1.0f;
					if(a == 1) pwr_scale = blkFx()->_chan[ch].out_pwr_scale;
					sgram.newData(_pix_bin[ch].getMap(plan), pwr_scale);
					update = true;
				}
			}

			if(update) sgram.blkDone(samp_pos, time_fft_n);
		}
		return;
	}

	// mono, input data always goes into upper sgram & output into lower sgram
	if(!_sgram[a].rdy()) return;

	_sgram[a].newData(_pix_bin[/*channel*/0].getMap(plan), /*scale*/1.0f);
	_sgram[a].blkDone(samp_pos, time_fft_n);
}

//-------------------------------------------------------------------------------------------------
bool Gui::keysRequired ()
{
	// don't expect keys for now
	//return false;

	if (frame && frame->getEditView()) return true;
	else return false;
}

//-------------------------------------------------------------------------------------------------
long Gui::onKeyDown (VstKeyCode &keyCode)
{
	if (frame && 0 && (keyCode.character >= 'a' && keyCode.character <= 'z'))
	{
		char val[64];
		char modifiers[32];
		strcpy (modifiers, "");
		if (keyCode.modifier & MODIFIER_SHIFT)
			strcpy (modifiers, "Shift+");
		if (keyCode.modifier & MODIFIER_ALTERNATE)
			strcat (modifiers, "Alt+");
		if (keyCode.modifier & MODIFIER_COMMAND)
			strcat (modifiers, "Cmd+");
		if (keyCode.modifier & MODIFIER_CONTROL)
			strcat (modifiers, "Ctrl+");

		sprintf (val, "onKeyDown : '%s%c'", modifiers, (char)(keyCode.character));
		return 1;
	}

	if (frame && (keyCode.virt == VKEY_UP || keyCode.virt == VKEY_DOWN))
	{
		CView *pView = frame->getCurrentView ();
	}
	return -1;
}

//-------------------------------------------------------------------------------------------------
long Gui::onKeyUp (VstKeyCode &keyCode)
{
	if (0 && (keyCode.character >= 'a' && keyCode.character <= 'z'))
	{
		char val[64];
		char modifiers[32];
		strcpy (modifiers, "");
		if (keyCode.modifier & MODIFIER_SHIFT)
			strcpy (modifiers, "Shift+");
		if (keyCode.modifier & MODIFIER_ALTERNATE)
			strcat (modifiers, "Alt+");
		if (keyCode.modifier & MODIFIER_COMMAND)
			strcat (modifiers, "Cmd+");
		if (keyCode.modifier & MODIFIER_CONTROL)
			strcat (modifiers, "Ctrl+");

		sprintf (val, "onKeyUp : '%s%c'", modifiers, (char)(keyCode.character));
		return 1;
	}

	return -1; 
}

//-------------------------------------------------------------------------------------------------
void Gui::resume ()
{
	// called when the plugin will be On
}

//-------------------------------------------------------------------------------------------------
void Gui::suspend ()
{
	// called when the plugin will be Off
	for(int i = 0; i < 2; i++) _sgram[i].suspend();
}

//-------------------------------------------------------------------------------------------------
void Gui::close ()
// virtual, override AEffGUIEditor::close()
// the GUI window is closing
{
	_ok = false;
	delete /*AEffGUIEditor::*/frame;
	frame = 0;
	// NOTE: don't delete any memory that FFTDataRdy might be using (i.e. anything to do with sgrams)
}

//-------------------------------------------------------------------------------------------------
void Gui::update(CDrawContext* context)
// virtual, override AEffGUIEditor
{
}

//-------------------------------------------------------------------------------------------------
void Gui::idle ()
// virtual, override AEffGUIEditor::idle()
{
	//DBG("Gui::idle");
	AEffGUIEditor::idle ();		// always call this to ensure update
}

//-------------------------------------------------------------------------------------------------
void Gui::setParameter (long index, float v)
// virtual, override AEffGUIEditor::setParameter()
//
// called from the AEffect to update the control's value
{
	using namespace BlkFxParam;
	BlkFxParam::SplitParamNum p(index);

	//DBG("setParameter" << VAR(index) << VAR(v) << VAR(frame) << VAR(p.fx_set) << VAR(p.glob_param));

	// test if the plug is opened
	if (!frame) return;

	// pass parameter
	if(p.fx_set >= 0 && p.fx_set < 16)
		_fx_ctrl[p.fx_set].setParameter(index, v);

	else if(p.glob_param >= 0)
		_glob_ctrl.setParameter(index, v);

	// call this to be sure that the graphic will be updated
	postUpdate ();
}
