/**************************************************************************************************
Copyright (c) 2005-2006 Darrell Tam
email: darrell.barrell@gmail.com

Spectrogram window
The spectrogram (waterfall) displays FFT data. Colour is used to represent magnitude of power
in FFT. Each block is put at the bottom of the display and the display scrolls upwards.

This class expects to be given a PixelFreqMap by the owning class in order to determine which
FFT bins correspond to which pixels

Bugs:
Currently doesn't scroll correctly if other windows are on top of it
In the buzz version it scrolled properly by calling the windows ScrollWindow function

History
	Date       Version    Programmer         Comments
	16/2/03    -         Darrell Tam		Buzz version
	01/12/05    1.0        Darrell Tam		Ported to VST


This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

***************************************************************************************************/
#include <dt/misc_stuff.h>
#include <dt/NoteFreq.h>
#include "Spectrogram.h"

#define DBG_FILE_NAME "c:\\sgram.txt"
#include "Debug.h"
using namespace std;

//-------------------------------------------------------------------------------------------------
Spectrogram::Spectrogram() :
	CViewContainer(
		CRect(0,0,0,0),		//const CRect &rect,
		NULL,				//CFrame *pParent
		NULL				//CBitmap *pBackground
	)
{
	DBG("Spectrogram" << VAR(this));
	name = "?";
	paused = false;

	tick_height = 1;
	frame_col = CColorRGB(0,0,0);

	// could try to get these from somewhere
	_pwr_min = 1e-4f * 1e-4f;
	_pwr_max = 0.2f * 0.2f;
	min_samps_per_line = 600;

	_ok = false;
}

Spectrogram::~Spectrogram()
{
	DBG("~Spectrogram" << VAR(this));
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::init(
	CFrame* parent, 
	CPoint p,					// window offset
	int height,
	PixelFreqMap* pix_freq		// pixel-frequency mapping (logarithmic)
)
//
// set the rectangle size, width is 2 + width of pixel freq map (borders)
//
// expect col map to be already set
{
	setParent(parent);

	/*CView::*/setViewSize(
		Offset(p, CRect(0, 0, pix_freq->numPixels()+2/*borders*/, height))
	);
	_pix_freq = pix_freq;
	_image.setNew(new CAlphaContext(parent, pix_freq->numPixels(), height+1));

	// fill image with opaque black
	CPoint pix;
	for(pix.y = 0; pix.y < height; pix.y++)
		for(pix.x = 0; pix.x < pix_freq->numPixels(); pix.x++) _image->drawPoint(pix, /*r*/0, /*g*/0, /*b*/0, /*a*/255);

	_vline_x[0] = _vline_x[1] = -1;
	_mouse.init();

	// clear out the image data
	_curr_max.resize(_image->getWidth());
	_curr_line_is_empty = true;


	// find client rectangle
	client_rect = size;
	client_rect.top++;
	client_rect.bottom -= /*Spectrogram::*/tick_height+1;
	client_rect.left++;
	client_rect.right = client_rect.left+_image->getWidth();

	// set initial position into buffer so that line 0 is displayed at the very top of the
	// client rectangle
	_disp_i[0] = _disp_i[1] = client_rect.height()-1;
	_image_i = _disp_i[0]+1;
	_curr_line_samp_pos = 0;

	// mouse info is drawn here
	info_rect = Offset(TopLeft(client_rect), CRect(5, 2, 120, 17));

	// msg/menu is drawn here
	menu_rect = Offset(TopRight(client_rect), CRect(-120, 2, 0, 17));

	_focus = false;

	_info_drawn = false;
	_menu_drawn = false;

	_info_dirty = true;
	_menu_dirty = true;

	// we're always dirty
	setDirty();

	if(_col_map) _ok = true;
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::suspend()
// call when sound suspends to reset sample position
{
	_curr_line_samp_pos = 0;
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::set(GradColMap& /*in*/ col_map)
{
	_col_map = &col_map;
	_col_log_offs = logf(_pwr_min);
	_col_log_scale = col_map.size()/(logf(_pwr_max)-_col_log_offs);
}

//-------------------------------------------------------------------------------------------------
inline unsigned long Spectrogram::mapCol(float in)
//
// return the colour corresponding to log of "in"
//
{
	GradColMap& col_map = *_col_map;
	unsigned out;
	if(in > _pwr_min) {
		out = (unsigned)((logf(in)-_col_log_offs)*_col_log_scale+0.5f);
		if(out >= _col_map->size()) out = col_map.size()-1;
	}
	else out = 0;
	return col_map[out];
}

//-------------------------------------------------------------------------------------------------
inline void Spectrogram::getMaxVals(
	pcplxf* bin_rng,	// appropriate bin rng's from PixelFreqMap
	bool empty_line,	// whether this is a new line (pass as constant to help optimizer)
	float pwr_scale		// pwr scale to apply
)
//
// get maximum values for each pixel from bin ranges
{
	//DBG("getMaxVals" << VAR(empty_line) << VAR(fft_idx));
	// loop over the pixels to find max value in each bin range

	float *curr_max = &_curr_max[0];
	float *curr_max_end = curr_max+_image->getWidth();

	// find max power in each pixel range
	pcplxf n = *bin_rng++;		// get first bin range
	while(curr_max != curr_max_end) {
		pcplxf r = n;		// current bin range
		n = *bin_rng++;			// next bin range

		// always do at least one bin for every output pixel
		float v = norm(*r++)*pwr_scale;
		float max_v;
		if(!empty_line) {
			max_v = *curr_max;
			if(v > max_v) max_v = v;
		}
		else {
			max_v = v;
		}

		// check any other bins up to the next bin position
		while(r.r < n.r) {
			float v = norm(*r++)*pwr_scale;
			if(v > max_v) max_v = v;
		}
		*curr_max++ = max_v;
	}
}

//------------------------------------------------------------------------------------------
void Spectrogram::newData(
	pcplxf* bin_rng,	// appropriate bin rng's from PixelFreqMap
	float pwr_scale		// scaling to apply to data
)
// add some new data to the spectrogram
{
//	DBG("newData" << VAR(samp_pos) << VAR(time_fft_n) << VAR(fft_idx));
	if(!_curr_line_is_empty) {
		getMaxVals(bin_rng, /*empty_line*/false, pwr_scale);
	}
	else {
		_curr_line_is_empty = false;
		getMaxVals(bin_rng, /*empty_line*/true, pwr_scale);
	}
}

//------------------------------------------------------------------------------------------
void Spectrogram::blkDone(
	long samp_pos,		// absolute sample position for start of this blk
	long time_fft_n		// number of samples in time domain (may be less samples than fft freq data)
)
{
	// don't display the line until we have processed enough samples
	if(samp_pos+time_fft_n-_curr_line_samp_pos < /*Spectrogram::*/min_samps_per_line) return;

	// render spectrogram line from the "curr_max" array into current line "_image_i"
	unsigned long *image_data = _image->data(/*x*/0, /*y*/_image_i);

	float *curr_max = &_curr_max[0];
	float *curr_max_end = curr_max+_image->getWidth();
	while(curr_max != curr_max_end)
		*image_data++ = mapCol(*curr_max++);

	// current line sample position
	_curr_line_samp_pos = samp_pos+time_fft_n;

	// this line is ready to be displayed
	_disp_i[1] = _image_i;

	// circular buffer
	_image_i++;
	if(_image_i >= _image->getHeight()) _image_i = 0;

	_curr_line_is_empty = true;
}


//-------------------------------------------------------------------------------------------------
void Spectrogram::drawBm(
	CDrawContext *context,	//
	CRect clip_r			// limit drawing to this clipping rectangle
)
//
// draw spectrogram bitmap
//
// use _disp_i[0] as the index of most recent line in image buffer
{
	//DBG("drawBm" << VAR(name) << VAR(clip_r));
	context->setDrawMode(kCopyMode);

	clip_r.bound(client_rect);

	// make sure that we aren't trying to draw more than what we have
	if(client_rect.bottom - clip_r.top > _image->getHeight()) {
		CRect r = clip_r;

		// adjust the output rectangle
		clip_r.top = client_rect.bottom - _image->getHeight();

		// fill rectangle at top for which there is no data with black
		r.bottom = clip_r.top;
		if(r.height() > 0) {
			context->setFillColor(kBlackCColor);
			context->fillRect(r);
		}
	}

	CRect r = clip_r;

	// perform copy in at most 2 blt's (depending on whether we need to wrap in src buffer)
	//
	int src_x = r.left-client_rect.left;

	while(1) {
		int n_lines = r.height();
		if(n_lines <= 0) break;

		// start line into image bitmap
		int src_i = 1+_disp_i[0] - (client_rect.bottom-r.top);
		bool temp = false;

		if(src_i < 0) {
			// wrap to end of source buffer
			src_i += _image->getHeight();

			// limit number of lines to end-of-buffer
			int max_lines = _image->getHeight()-src_i;
			if(max_lines < n_lines) n_lines = max_lines;
			temp = true;
		}

		// do the copy
		CRect dst_r = r;
		dst_r.bottom = dst_r.top + n_lines;
		CRect src_r(src_x, src_i, src_x+dst_r.width(), src_i+n_lines);
		_image->blend(context, dst_r, src_r, /*overall alpha*/255, /*src alpha*/false);

		r.top = dst_r.bottom;
	}
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::invertVLine(CDrawContext *context)
// draw/erase the vertical line
{
	CRect r = client_rect;
	r.left += _vline_x[0];
	r.right = r.left;
	r.bottom--;
	FullInvertRect(context, r);
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::eraseOverlay(CDrawContext *context)
// erase whatever overlay stuff is dirty
{
	//DBG("eraseOverlay" << VAR(name) << VAR(_info_drawn) << VAR(_info_dirty) << VAR(_menu_drawn) << VAR(_menu_dirty));
	// erase the line if we're going to update to anything
	if((_vline_x[0] >= 0) && (_info_dirty || _menu_dirty)) {
		invertVLine(context);
		_vline_x[0] = -1;
	}

	if(_info_drawn && _info_dirty) {
		drawBm(context, info_rect);
		_info_drawn = false;
	}

	if(_menu_drawn && _menu_dirty) {
		drawBm(context, menu_rect);
		_menu_drawn = false;
	}
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::drawOverlay(CDrawContext *context)
// draw whatever overlay stuff is dirty (require that background has already been erased)
// clear dirty flags
{
	//DBG("drawOverlay" << VAR(name) << VAR(_info_drawn) << VAR(_info_dirty) << VAR(_menu_drawn) << VAR(_menu_dirty));

	// update line position after erasing
	_vline_x[0] = _vline_x[1];

	// prepare to write some text
	context->setFont(kNormalFont, /*sz*/0, kBoldFace);
	context->setFillColor(CColorRGB(255,255,255));
	context->setFontColor(CColorRGB(0,0,0));

	// draw the vline info if need be
	if(_vline_x[0] < 0) _info_dirty = false;
	if(_info_dirty) {
		char str[256], *s;
		float hz = _pix_freq->pixToFreq(_vline_x[0]);
		s = str;
		s += HzToNote(s, hz);
		OutlineDrawString(
			context,		// draw context
			LeftHalf(/*Spectrogram::*/info_rect),// rectangle
			str,			// string
			kCenterText		// horiz alignment
		);

		s = str;
		s += sprintNum(s, hz);
		s += sprintf(s, "Hz");
		OutlineDrawString(
			context,		// draw context
			RightHalf(/*Spectrogram::*/info_rect),// rectangle
			str,			// string
			kCenterText		// horiz alignment
		);
		_info_drawn = true;
	}

	// draw the msg if need be
	if(_menu_dirty) {
		char str_[128];
		const char* str;

		// get the string out of the menu
		if(paused) str = "paused";
		else if(menu) {
			str_[0] = 0; // empty string
			menu->getCurrent(str_);
			menu->setDirty(false);
			str = str_;
		}
		// no menu attached, use name
		else str = name;

		// draw some text
		OutlineDrawString(
			context,		// draw context
			menu_rect,		// rectangle
			str,			// string
			kCenterText		// horiz alignment
		);
		_menu_drawn = true;
	}

	// draw the line
	if(_vline_x[0] >= 0)
		invertVLine(context);

	_menu_dirty = false;
	_info_dirty = false;
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::update(CDrawContext *context)
// virtual, override CViewContainer::update
//
// called periodically, performs scrolling if need be and displays mouse information
{
	_mouse.update(context);
	if(menu && _mouse.release_mask) paused = false;

	if(_mouse.IsInside(size)) {
		// mouse is inside our rectangle
		setVLinePix(_mouse.pos.x-client_rect.left);
		_focus = true;
	}
	else if(_focus) {
		// mouse has gone out of focus
		clrVLine();
		_focus = false;
	}

	// work out the number of lines to scroll
	int new_disp_i = _disp_i[1];
	// note: _disp_i[1] may be updated asynchronously to this function - ensure code is multi-thread safe
	int scroll_n = new_disp_i-_disp_i[0];
	if(scroll_n < 0) scroll_n += _image->getHeight();

	// if not scrolling: redraw overlay if need be
	if(!scroll_n) {
		// check what things need drawing
		if(menu) {
			if(menu->isDirty()) _menu_dirty = true;
		}
		if(_vline_x[0] != _vline_x[1]) _info_dirty = true;
		if(!_menu_dirty && !_info_dirty) return;
		//DBG("update" << VAR(name) << VAR(scroll_n));
		eraseOverlay(context);
		drawOverlay(context);
		return;
	}

	// scroll upwards
	//DBG("update" << VAR(name) << VAR(scroll_n));

	// force removal of overlay stuff
	_menu_dirty = true;
	_info_dirty = true;
	eraseOverlay(context);

	// update the index to most recent line
	_disp_i[0] = new_disp_i;

	// TODO: someone port this part to mac
	HWND wnd = (HWND)(getParent()->getSystemWindow());
 	RECT r = {client_rect.left, client_rect.top, client_rect.right, client_rect.bottom };
	ScrollWindow(wnd, 0, -scroll_n, /*scroll rect*/&r, /*clip rect*/&r);

	// force repaint of missing part synchronously
	UpdateWindow((HWND)getParent()->getSystemWindow());

	// put on the overlay
	drawOverlay(context);
}


//-------------------------------------------------------------------------------------------------
void Spectrogram::drawRect(CDrawContext *context, CRect& update_rect)
// virtual override CViewContainer::drawRect
{
	//DBG("drawRect" << VAR(name) << VAR(client_rect) << VAR(update_rect));

	if(update_rect.left < client_rect.left || update_rect.right > client_rect.right ||
	update_rect.top < client_rect.top || update_rect.bottom > client_rect.bottom) {
		// draw borders and ticks if update is outside of client rect

		context->setFillColor(/*Spectrogram::*/frame_col);
		// top border
		CRect r = /*CView::*/size;
		r.right--;
		r.bottom = client_rect.top-1;
		FullFillRect(context, r);
		// bottom border
		r.top = client_rect.bottom;
		r.bottom = r.top;
		FullFillRect(context, r);
		// left edge
		r.top = size.top;
		r.bottom = client_rect.bottom;
		r.left = size.left;
		r.right = client_rect.left-1;
		FullFillRect(context, r);
		// right edge
		r.left = client_rect.right;
		r.right = size.right-1;
		FullFillRect(context, r);
		// octave ticks
		r.top = client_rect.bottom+1;
		r.bottom = size.bottom-1;
		for(float x=0; x < size.width(); x += _pix_freq->pixPerOctave()) {
			r.left = (int)(x+0.5f)+client_rect.left;
			r.right = r.left;
			FullFillRect(context, r);
		}
	}

	// force redraw of overlay graphics	
	_menu_dirty = true;
	_info_dirty = true;

	// draw the bitmap & let "update" put on the overlay
	drawBm(context, update_rect);
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::draw(CDrawContext *context)
// virtual override CViewContainer::draw
//
{
	// redraw everything
	drawRect(context, size);
}

//-------------------------------------------------------------------------------------------------
void Spectrogram::mouse(CDrawContext *context, CPoint &where)
// virtual, override CView method
//
// corresponds to mouse button press
//
{
	_mouse.click(context, where, size);

	// if there's a menu give the mouse button to it
	if(menu) {
		if(_mouse.click_mask) paused = true;
		menu->mouse(context, /*CPoint*/_mouse);
	}

	// otherwise default behaviour is to toggle pause
	else if(_mouse.click_mask) {
		paused = !paused;
		_menu_dirty = true;
	}
}

//-------------------------------------------------------------------------------------------------
bool Spectrogram::removed(CView* parent)
// virtual, override CViewContainer::removed
{
	DBG("removed");
	CViewContainer::removed(parent);

	// could do more cleanup here
	_ok = false;
	setParent(NULL);
	return true;
}
